/*
 *  linux/fs/beos/file.c
 *
 * Copyright (C) 1999  Makoto Kato (m_kato@ga2.so-net.ne.jp)
 *
 *  from
 *
 *  linux/fs/minix/file.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 */

#include <asm/uaccess.h>
#include <asm/system.h>

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/beos_fs.h>
#include <linux/fcntl.h>
#include <linux/sched.h>
#include <linux/stat.h>
#include <linux/locks.h>
#include <linux/mm.h>
#include <linux/pagemap.h>


ssize_t beos_file_read (struct file *, char *,  size_t, loff_t *);
static beos_inode_addr beos_startpos_from_ds (struct super_block *,
	beos_data_stream *, off_t);
static int beos_read_indirect_block (struct super_block *, const beos_inode_addr,
	const int, beos_inode_addr *);
static int beos_read_double_indirect_block (struct super_block *,
	const beos_inode_addr, const int, beos_inode_addr *);
static ssize_t beos_file_write (struct file *, const char *, size_t, loff_t *);


static struct file_operations beos_file_ops =
{
	NULL,			/* lseek */
	beos_file_read,
#ifdef CONFIG_BEOS_FS_RW
	beos_file_write,
#else
	NULL,
#endif
	NULL,			/* readdir - bad */
	NULL,			/* poll - default */
	NULL,			/* ioctl - default */
	generic_file_mmap,
	NULL,
	NULL,			/* flush */
	NULL,			/* release */
	file_fsync,
};

struct inode_operations beos_file_inode_operations =
{
	&beos_file_ops,		/* default file operations */
	NULL,					/* create */
	NULL,					/* lookup */
	NULL,					/* link */
	NULL,					/* unlink */
	NULL,					/* symlink */
	NULL,					/* mkdir */
	NULL,					/* rmdir */
	NULL,					/* mknod */
	NULL,					/* rename */
	NULL,					/* readlink */
	NULL,					/* follow_link */
	generic_readpage,	/* readpage */
	NULL,					/* writepage */
	&beos_bmap,		/* bmap */
	NULL,					/* truncate */
	NULL,					/* permission */
	NULL					/* smap */
};


ssize_t beos_file_read(struct file *filp, char *buf,  size_t count, loff_t *ppos)
{
	struct inode *       inode = filp->f_dentry->d_inode;
	struct super_block * sb = inode->i_sb;
	beos_data_stream *    ds = &inode->u.beos_i.i_data.ds;
	beos_inode_addr       iaddr;
	beos_inode_addr       iaddr2;
	off_t                pos = *ppos;
	off_t                sum = 0;
	int                  i;
	int                  j;
	off_t                read_count;
	off_t                offset;

	BEOS_OUTPUT (("---> beos_file_read() "
		"inode %lu count %lu ppos %Lu\n",
		inode->i_ino, (__u32) count, *ppos));

	/*
	 * Get start position
	 */

	for (i = 0; ;) {
		iaddr = beos_read_data_stream (sb, ds, &i);
		if (BEOS_IS_EMPTY_IADDR(&iaddr)) {
			BEOS_OUTPUT ((" end data_stream\n"));
			break;
		}

		if ((pos >= sum)
			&& (pos < sum + iaddr.len * sb->u.beos_sb.block_size)) {

			iaddr.start += (pos - sum) / sb->u.beos_sb.block_size;
			iaddr.len -= (pos - sum) / sb->u.beos_sb.block_size;
			break;
		}
		sum += iaddr.len * sb->u.beos_sb.block_size;
	}

	read_count = 0;
	offset = (pos - sum) % sb->u.beos_sb.block_size;
	iaddr2 = iaddr;

	count = inode->i_size > pos + count ? count : inode->i_size - pos;

	for ( j = 0; j < iaddr.len; j++) {

		struct buffer_head * bh;

		bh = beos_bread2 (sb, iaddr2);
		if (!bh)
			return -EBADF;

		BEOS_OUTPUT ((" read_count %Ld offset %Ld\n",
			read_count, offset));

		if (count > sb->u.beos_sb.block_size) {

			memcpy (buf + read_count, bh->b_data + offset,
				sb->u.beos_sb.block_size - offset);

			*ppos += sb->u.beos_sb.block_size - offset;
			count -= sb->u.beos_sb.block_size - offset;
			read_count += sb->u.beos_sb.block_size - offset;
			offset = 0;

			iaddr2.start += 1;

			brelse (bh);

		} else {

			memcpy (buf + read_count, bh->b_data + offset, count);

			*ppos += count;
			read_count += count;
	
			brelse (bh);
			break;
		}
	}

	ppos += read_count;

	BEOS_OUTPUT (("<--- beos_file_read() "
		"return value %d, ppos %Ld\n", read_count, *ppos));

	return read_count;
}

/*
 * Read indirect block
 */

static int beos_read_indirect_block (struct super_block * sb,
	const beos_inode_addr indirect, const int pos, 
	beos_inode_addr * iaddr)
{
	beos_inode_addr *     array;
	struct buffer_head * bh;
	beos_inode_addr       addr = indirect;
	int                  p = pos;

	BEOS_OUTPUT (("---> beos_read_indirect_block()\n"));

	/*
	 * explore nessealy block
	 */

	addr.start += p / BEOS_BLOCK_PER_INODE(sb);
	addr.len -= p / BEOS_BLOCK_PER_INODE(sb);

	if (addr.len < 1) {
		goto error;
	}

	p = p % BEOS_BLOCK_PER_INODE(sb);

	bh = beos_bread2 (sb, addr);
	if (!bh) {
		goto error;
	}

	array = (beos_inode_addr *) bh->b_data;

	/*
	 * Is this block inode address??
	 */

	if (!BEOS_IS_EMPTY_IADDR(&array[p % BEOS_BLOCK_PER_INODE(sb)])) {
		*iaddr = array[p % BEOS_BLOCK_PER_INODE(sb)];
	} else {
		brelse (bh);
		goto error;
	}

	brelse (bh);

	BEOS_OUTPUT (("<--- beos_read_indirect_block()\n"));

	return 0;

error:
	printk ("BEOS: cannot read indirect block\n");
	return -EBADF;
}

/*
 * beos_read_double_indirect_block
 *
 * description:
 *  Read double-indirect block
 *
 * parameter:
 *  sb              ... super block
 *  double_indirect ... inode address of double-indirect block
 *  pos             ... position of double-indirect block
 *                       (*pos - max_direct_block - max_indirect_block)
 *
 * return value:
 *  0 ... sucess
 */

static int beos_read_double_indirect_block (struct super_block * sb,
        const beos_inode_addr double_indirect, const int pos,
	beos_inode_addr * iaddr)
{
        struct buffer_head *  bh_indirect;
        int                   p = pos;
        beos_inode_addr *      indirects;
	beos_inode_addr        addr = double_indirect;
	int                   i;

	BEOS_OUTPUT (("---> beos_read_double_indirect_block() \n"));

        while (addr.len) {
                bh_indirect = beos_bread2 (sb, addr);

                if (!bh_indirect) {
                        BEOS_OUTPUT (("cannot read double-indirect block "
                                "[%lu, %u, %u]\n",
                                double_indirect.allocation_group,
                                double_indirect.start, double_indirect.len));
                        return -EBADF;
                }

                addr.start++;
                addr.len--;

                indirects = (beos_inode_addr *) bh_indirect->b_data;

                for (i = 0; i < BEOS_BLOCK_PER_INODE(sb); i++) {

                        if (p < indirects[i].len * BEOS_BLOCK_PER_INODE(sb)) {

                                /*
                                 * find block!
                                 */

				int err;

				err = beos_read_indirect_block (sb,
					indirects[i], p, iaddr);
				brelse (bh_indirect);

				return err;
                        }

                        p -= indirects[i].len * BEOS_BLOCK_PER_INODE(sb);
                }

                brelse(bh_indirect);
        }

        /*
         * cannot find block...
         */

        BEOS_OUTPUT (("cannot find block...\n"));

        return -EIO;
}


beos_inode_addr beos_read_data_stream (struct super_block * sb,
	beos_data_stream * ds, int * pos)
{
        beos_inode_addr       iaddr = {0, 0, 0};

	BEOS_OUTPUT (("---> beos_read_data_stream() "
		"pos %d\n", *pos));

        if( *pos < 0 )
                return iaddr;

        if (*pos < ds->max_direct_range) {

		/*
		 * This position is in direct block.
		 */

		BEOS_OUTPUT ((" read in direct block [%lu, %u, %u]\n",
			ds->direct[*pos].allocation_group,
			ds->direct[*pos].start, ds->direct[*pos].len));

		/*
		 * Is this block inode address??
		 */

                if (!BEOS_IS_EMPTY_IADDR(&ds->direct[*pos])) {
			iaddr = ds->direct[(*pos)++];
		}
        } else if (*pos < ds->max_indirect_range + ds->max_direct_range) {

		/*
		 * This position is in in-direct block.
		 */

		int  p = *pos - ds->max_direct_range;
		
		BEOS_OUTPUT ((" read in indirect block [%lu, %u, %u]\n",
			ds->indirect.allocation_group, ds->indirect.start,
			ds->indirect.len));

		if (!beos_read_indirect_block (sb, ds->indirect, p, &iaddr)) {
			(*pos)++;
		} 
        } else if (*pos < ds->max_double_indirect_range
		+ ds->max_direct_range + ds->max_indirect_range) {

		/*
		 * This position is in double-in-direct block.
		 */

		int  p = *pos - ds->max_direct_range - ds->max_indirect_range;

		BEOS_OUTPUT ((" read in double indirect block\n"));

		if (!beos_read_double_indirect_block (sb, ds->double_indirect,
			p, &iaddr)) {

			(*pos)++;
		} 
        } else {
		/*
		 * out of range block
		 */
	}

	if (BEOS_IS_EMPTY_IADDR(&iaddr))
		*pos = 0;

        return iaddr;
}

/*
 * Get inode address of start position form data stream
 */

static beos_inode_addr beos_startpos_from_ds (struct super_block * sb,
					beos_data_stream * ds, off_t pos)
{
	beos_inode_addr iaddr;
	off_t          sum = 0;
	int            i = 0;

	BEOS_OUTPUT (("---> beos_startpos_from_ds()\n"));

	/*
	 * Get start position
	 */

	while(1) {
		iaddr = beos_read_data_stream (sb, ds, &i);
		if (BEOS_IS_EMPTY_IADDR(&iaddr)) {
			BEOS_OUTPUT ((" end data_stream\n"));
			break;
		}

		if ((pos >= sum)
			&& (pos < sum + iaddr.len * sb->u.beos_sb.block_size)) {
			iaddr.start += (pos - sum) / sb->u.beos_sb.block_size;
			iaddr.len -= (pos - sum) / sb->u.beos_sb.block_size;
			break;
		}
		sum += iaddr.len * sb->u.beos_sb.block_size;
	}

	BEOS_OUTPUT (("<--- beos_startpos_from_ds()\n"));

	return iaddr;
}

#ifdef CONFIG_BEOS_RW 
static ssize_t beos_file_write (struct file * filp, const char * buf,
                                size_t count, loff_t *ppos)
{
	struct inode *       inode = filp->f_dentry->d_inode;
	struct super_block * sb;
//	struct buffer_head * bh;
	beos_data_stream *    ds;
	beos_inode_addr       iaddr;
	loff_t               pos;
	ssize_t              written;

	if (!inode) {
		return -EINVAL;
	}

	if (!S_ISREG(inode->i_mode)) {
		return -EINVAL;
	}

	sb = inode->i_sb;
	ds = &inode->u.beos_i.i_data.ds;

	if (filp->f_flags & O_APPEND)
		pos = inode->i_size;
	else
		pos = *ppos;

	iaddr = beos_startpos_from_ds (sb, ds, pos);
	written = 0;

	if (pos > inode->i_size)
		inode->i_size = pos;

	inode->i_mtime = inode->i_ctime = CURRENT_TIME;
	*ppos = pos;
	mark_inode_dirty(inode);

	return written;
}
#endif
