/*
 *  linux/fs/beos/namei.c
 *
 * Copyright (C) 1999  Makoto Kato (m_kato@ga2.so-net.ne.jp)
 *
 *  from
 *
 *  linux/fs/minix/namei.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 */

#include <asm/uaccess.h>

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/fcntl.h>
#include <linux/sched.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <linux/locks.h>
#include <linux/quotaops.h>


/*
 * beos_find_entry
 *
 * description:
 *  find file or directory entry
 *
 * parameter:
 *  dir     ... parent's inode
 *  name    ... file or directry name
 *  namelen ... length of name
 */

static beos_off_t beos_find_entry (struct inode * dir, const char * const name,
	int namelen)
{
	struct buffer_head *    bh;
	beos_index_node *	bn;
	struct super_block *	sb = dir->i_sb;
	int			sd_pos = 0;
	int			key_pos;
	char			key[BEOS_NAME_LEN + 1];
	int			key_len;
	beos_off_t		value;

	// BEOS_OUTPUT (
	printk("BEOS ---> beos_find_entry()\n");

	while (sd_pos >= 0)
	{
		beos_data_stream *	ds = &dir->u.beos_i.i_data.ds;
		beos_inode_addr		iaddr;
		int			flags;
		beos_off_t		offset;

		/*
		 * get index node from small_data.
		 *
		 * First data stream has index entry, but other than it,
		 * doesn't have it.  
		 */

		if (sd_pos)
			flags = 1;
		else
			flags = 0;

		iaddr = beos_read_data_stream (sb, ds, &sd_pos);
		if (sd_pos <= 0)
		{
			/*
			 * cannot read next data stream
			 */

			//BEOS_OUTPUT (
			printk(
			"BEOS <--- beos_find_entry() cannot read next data stream\n");
			return 0;
		}

		bh = beos_read_index_node (iaddr, sb, flags, &offset);
		if (!bh)
		{
			/*
			 * cannot read index node
			 */

			//BEOS_OUTPUT (
			printk(
			"BEOS <--- beos_find_entry() cannot read index node\n");
			return 0;
		}

		bn = (beos_index_node *) (bh->b_data + offset);

		/*
		 * explore in this block
		 */

		key_pos = 0;
		while (key_pos >= 0)
		{
			beos_get_key_from_index_node (bn, &key_pos,
				sb->u.beos_sb.mount_opts.beos_type, key,
				&key_len, &value);
			if (key_pos < 0)
			{
				//BEOS_OUTPUT (
				printk("BEOS: key not exist in this block\n");
				break;
			}

			if (!strcmp (key, name))
			{
				brelse (bh);
				//BEOS_OUTPUT (
				printk(
				"BEOS <--- beos_find_entry() value = %Lu\n", value);
				return value;
			}
		}

		brelse (bh);
	}

	//BEOS_OUTPUT (
	printk("BEOS <--- beos_find_entry() not found\n");
	return 0;
}


#if 0
struct dentry *beos_lookup(struct inode * dir, struct dentry *dentry)
{
	struct inode *		inode = NULL;
	beos_off_t		offset;
	unsigned int		len = dentry->d_name.len;
	int			olen;
	const unsigned char *	tmpname = dentry->d_name.name;
	//char			tmpname2;
	char *			tmpname2;
	struct super_block *	sb = dir->i_sb;

	//BEOS_OUTPUT (
	printk(
	"BEOS ---> beos_lookup() name '%s' index inode '%ld'\n", tmpname, dir->i_ino);

	/*
	 * Convert to UTF-8
	 */
	// char * beos_nls2utf (unsigned char * src, int srclen, char ** dist,
	// int * len, struct super_block * sb)

	if (!beos_nls2utf (*tmpname, len, &tmpname2, &olen, sb))
	{
		printk(
		KERN_ERR "BEOS beos_lookup() unable to allocate memory\n");
		return ERR_PTR(-ENOMEM);
	}

	if (olen > BEOS_NAME_LEN)
	{
		printk(
		KERN_ERR "BEOS beos_lookup() name too long: '\"%s\"'",tmpname2);
		putname (tmpname2);
		return ERR_PTR(-ENAMETOOLONG);
	}

	offset = beos_find_entry (dir, tmpname2, olen);
	putname (tmpname2);

	if (offset)
	{
		inode = iget (sb, (ino_t) offset);
		if (!inode)
		{
			return ERR_PTR(-EACCES);
		}
	}

	d_add (dentry, inode);

	//BEOS_OUTPUT (
	printk("BEOS <--- beos_lookup()\n");

	return;
}
#endif


struct dentry *beos_lookup (struct inode * dir, struct dentry *dentry)
{
	struct inode *		inode = NULL;
	beos_off_t		offset;
	unsigned int		len = dentry->d_name.len;
	int			olen;
	//const unsigned char *	tmpname = dentry->d_name.name;
	char *			tmpname = dentry->d_name.name;
	char *			tmpname2;
	struct super_block *	sb = dir->i_sb;

	//BEOS_OUTPUT (
	printk(
	"BEOS ---> beos_lookup() name '%s' index inode '%ld'\n", tmpname, dir->i_ino);

	/*
	 * Convert to UTF-8
	 */

	if (!beos_nls2utf (tmpname, len, &tmpname2, &olen, sb))
	{
		return ERR_PTR(-ENOMEM);
	}

	if (olen > BEOS_NAME_LEN)
	{
		printk("BEOS beos_lookup() name too long: '\"%s\"'",tmpname2);
		putname (tmpname2);
		return ERR_PTR(-ENAMETOOLONG);
	}

	printk("BEOS beos_lookup() name len %i\n",olen);

	offset = beos_find_entry (dir, tmpname2, olen);
	putname (tmpname2);

	if (offset)
	{
		inode = iget (sb, (ino_t) offset);
		if (!inode)
		{
			return ERR_PTR(-EACCES);
		}
	}

	d_add (dentry, inode);

	//BEOS_OUTPUT (
	printk("BEOS <--- beos_lookup()\n");

	return NULL;
}


#if 0
int beos_lookup (struct inode * dir, struct dentry *dentry)
{
	struct inode *		inode = NULL;
	beos_off_t		offset;
	unsigned int		len = dentry->d_name.len;
	int			olen;
	//const unsigned char *	tmpname = dentry->d_name.name;
	char *			tmpname = dentry->d_name.name;
	char *			tmpname2;
	struct super_block *	sb = dir->i_sb;

	//BEOS_OUTPUT (
	printk(
	"BEOS ---> beos_lookup() name '%s' index inode '%ld'\n", tmpname, dir->i_ino);

	/*
	 * Convert to UTF-8
	 */

	if (!beos_nls2utf (tmpname, len, &tmpname2, &olen, sb))
	{
		return -ENOMEM;
	}

	if (olen > BEOS_NAME_LEN)
	{
		printk("BEOS beos_lookup() name too long: '\"%s\"'",tmpname2);
		putname (tmpname2);
		return -ENAMETOOLONG;
	}

	printk("BEOS beos_lookup() name len %i\n",olen);

	offset = beos_find_entry (dir, tmpname2, olen);
	putname (tmpname2);

	if (offset)
	{
		inode = iget (sb, (ino_t) offset);
		if (!inode)
		{
			return -EACCES;
		}
	}

	d_add (dentry, inode);

	//BEOS_OUTPUT (
	printk("BEOS <--- beos_lookup()\n");

	return 0;
}
#endif
